# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("EMR::Studio")
class EMRStudioScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        studios = item.get("Studios", [])
        for studio in studios:
            studio_id = studio.get("StudioId", None)
            if studio_id is None:
                continue

            detail = client.describe_studio(StudioId=studio_id).get("Studio", {})
            if overwrite or not self._has_tag(detail, key):
                resources.append((detail.get("StudioArn"), self._get_tag(detail, key)))

        return resources

    @property
    def _client_name(self) -> str:
        return "emr"

    @property
    def _paginator(self):
        return "list_studios"

    @property
    def _filters(self):
        return {}



# -*- coding: utf-8 -*-

from awstt.worker.scanner import Scanner
from awstt.worker.scanners.ami import AMIScanner
from awstt.worker.scanners.autoscaling import AutoScalingScanner
from awstt.worker.scanners.cloudformation import CloudFormationScanner
from awstt.worker.scanners.cloudfront import CloudFrontScanner
from awstt.worker.scanners.cloudtrail import CloudTrailScanner
from awstt.worker.scanners.cloudwatch_alarm import CloudWatchAlarmScanner
from awstt.worker.scanners.cloudwatch_rules import CloudWatchRulesScanner
from awstt.worker.scanners.dynamodb import DynamoDBScanner
from awstt.worker.scanners.ebs import EBSScanner
from awstt.worker.scanners.ec2 import EC2Scanner
from awstt.worker.scanners.efs import EFSScanner
from awstt.worker.scanners.eip import EIPScanner
from awstt.worker.scanners.elasticache_cluster import ElastiCacheClusterScanner
from awstt.worker.scanners.elasticache_snapshot import ElastiCacheSnapshotScanner
from awstt.worker.scanners.elasticsearch import ElasticsearchScanner
from awstt.worker.scanners.elb import ELBScanner
from awstt.worker.scanners.emr_cluster import EMRClusterScanner
from awstt.worker.scanners.emr_studio import EMRStudioScanner
from awstt.worker.scanners.eni import ENIScanner
from awstt.worker.scanners.iam_policy import IAMPolicyScanner
from awstt.worker.scanners.iam_role import IAMRoleScanner
from awstt.worker.scanners.iam_user import IAMUserScanner
from awstt.worker.scanners.internet_gateway import InternetGatewayScanner
from awstt.worker.scanners.kms import KMSScanner
from awstt.worker.scanners.lambda_function import LambdaScanner
from awstt.worker.scanners.lightsail import LightsailScanner
from awstt.worker.scanners.nat_gateway import NATGatewayScanner
from awstt.worker.scanners.network_acl import NetworkACLScanner
from awstt.worker.scanners.opensearch import OpenSearchScanner
from awstt.worker.scanners.rds_cluster import RDSClusterScanner
from awstt.worker.scanners.rds_instance import RDSInstanceScanner
from awstt.worker.scanners.routetable import RouteTableScanner
from awstt.worker.scanners.s3_bucket import S3BucketScanner
from awstt.worker.scanners.security_group import SecurityGroupScanner
from awstt.worker.scanners.snapshot import SnapshotScanner
from awstt.worker.scanners.subnet import SubnetScanner
from awstt.worker.scanners.vpc import VPCScanner
from awstt.worker.scanners.vpc_peering import VPCPeeringScanner
from awstt.worker.scanners.vpn_gateway import VPNGatewayScanner
from awstt.worker.tagger import Tagger
from awstt.worker.utils import detect_account_id, detect_partition, detect_regions



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Optional, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("IAM::Role")
class IAMRoleScanner(Scanner):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        account_id: str,
        *,
        partition: str = "aws",
        regions: Optional[List[str]] = None,
        profile: Optional[str] = None,
    ):
        """
        IAM is global service, ensure just create one global client
        """
        super().__init__(account_id, profile=profile, regions=["global"])

    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        roles = item.get("Roles", [])
        for role in roles:
            resource_tags = {self._tags_key: []}
            tags_paginator = client.get_paginator("list_role_tags").paginate(RoleName=role.get("RoleName"))
            # noinspection DuplicatedCode
            for page in tags_paginator:
                resource_tags[self._tags_key].extend(page.get(self._tags_key, []))

            if overwrite is True or not self._has_tag(resource_tags, key):
                resources.append((role.get("Arn"), self._get_tag(resource_tags, key)))

        return resources

    @property
    def _client_name(self) -> str:
        return "iam"

    @property
    def _paginator(self):
        return "list_roles"

    @property
    def _filters(self):
        return {}



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Optional, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("Lambda::Function")
class LambdaScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        for func in item.get("Functions", []):
            arn = func.get("FunctionArn", None)
            if arn is None:
                continue

            resource_tags = client.list_tags(Resource=arn).get("Tags", {})
            origin_value = self._get_tag(resource_tags, key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    def _has_tag(self, item: dict, key: str) -> bool:
        return key in item

    def _get_tag(self, item: dict, key: str) -> Optional[str]:
        return item.get(key, None)

    @property
    def _client_name(self) -> str:
        return "lambda"

    @property
    def _paginator(self):
        return "list_functions"

    @property
    def _filters(self):
        return {}

    @property
    def _tags_key(self) -> str:
        return "Items"



# -*- coding: utf-8 -*-

import sys
import textwrap
from argparse import SUPPRESS, ArgumentParser, RawDescriptionHelpFormatter, RawTextHelpFormatter

from awstt.executor import execute
from awstt.worker import *


class RawFormatter(RawTextHelpFormatter, RawDescriptionHelpFormatter):
    """Formats argument help which maintains line length restrictions as well as appends default value if present."""

    def _split_lines(self, text, width):
        wrapper = textwrap.TextWrapper(width=width)
        lines = []
        for line in text.splitlines():
            if len(line) > width:
                lines.extend(wrapper.wrap(line))
            else:
                lines.append(line)
        return lines


def run():
    parser = ArgumentParser(
        prog="awstt",
        description="AWS-Tag-Tools: A bulk management tool for the tags of AWS resources",
        formatter_class=RawFormatter,
        usage=SUPPRESS,
    )

    parser.add_argument(
        "--key",
        type=str,
        required=False,
        help="the key of tag will be tagged to resources",
    )
    parser.add_argument(
        "--value",
        type=str,
        required=False,
        help="the value of tag will be tagged to resources",
    )
    parser.add_argument(
        "--overwrite",
        required=False,
        default=False,
        action="store_true",
        help="whether to overwrite exists value of tag when key is existed\ndefault to `False`",
    )
    parser.add_argument(
        "--regions",
        type=str,
        required=False,
        help="the AWS regions to execute actions\nwill auto detect if not set\nuse `','` to separate multi regions",
    )
    parser.add_argument(
        "--profile",
        type=str,
        required=False,
        help="the name of AWS credentials profile to use",
    )
    parser.add_argument(
        "--partition",
        type=str,
        required=False,
        choices=["aws", "aws-cn", "aws-us-gov"],
        default="aws",
        help="the partition to execute actions, must be one of `'aws' | 'aws-cn' | 'aws-us-gov'`\ndefault to `'aws'`",
    )
    parser.add_argument(
        "--list-services",
        required=False,
        action="store_true",
        help="list all supported services by this tool and exit",
    )

    args = parser.parse_args()

    if args.list_services is True:
        print("==== supported services ====")
        workers = Scanner.list_available()
        for worker in workers:
            print(f" --> {worker}")

        sys.exit(0)

    if args.key is None or args.value is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    execute(
        dict(
            partition=args.partition,
            regions=args.regions,
            profile=args.profile,
            tag_key=args.key,
            tag_value=args.value,
            overwrite=args.overwrite,
        ),
        None,
    )



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("ElastiCache::Snapshot")
class ElastiCacheSnapshotScanner(Scanner):
    # noinspection PyUnusedLocal
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        snapshots = item.get("Snapshots", [])
        # noinspection DuplicatedCode
        for snapshot in snapshots:
            arn = snapshot.get("ARN", None)
            if arn is None:
                continue

            resource_tags = client.list_tags_for_resource(ResourceName=arn)
            if overwrite or not self._has_tag(resource_tags, key):
                resources.append((arn, self._get_tag(resource_tags, key)))

        return resources

    @property
    def _client_name(self) -> str:
        return "elasticache"

    @property
    def _paginator(self):
        return "describe_snapshots"

    @property
    def _arn_resource_type(self) -> str:
        return "snapshot"

    @property
    def _tags_key(self) -> str:
        return "TagList"



# -*- coding: utf-8 -*-

import logging
import os


def init_logger():
    if os.getenv("LAMBDA_TASK_ROOT") is None:
        import inspect
        import sys

        from loguru import logger

        class InterceptHandler(logging.Handler):
            def emit(self, record: logging.LogRecord) -> None:
                # Get corresponding Loguru level if it exists.
                level: str | int
                try:
                    level = logger.level(record.levelname).name
                except ValueError:
                    level = record.levelno

                # Find caller from where originated the logged message.
                frame, depth = inspect.currentframe(), 0
                while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
                    frame = frame.f_back
                    depth += 1

                logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

        logger.configure(
            handlers=[
                {
                    "sink": sys.stderr,
                    "format": (
                        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
                        "<level>{level: <8}</level> | "
                        "<cyan>{name}:{line}</cyan> - <level>{message}</level>"
                    ),
                }
            ]
        )
        logging.getLogger().handlers = [InterceptHandler()]
    else:
        import boto3

        boto3.set_stream_logger("", logging.INFO)

    logging.getLogger().setLevel("INFO")
    logging.getLogger("botocore").setLevel(logging.ERROR)



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("ELB")
class ELBScanner(Scanner):
    # noinspection PyUnusedLocal
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        balancers = item.get("LoadBalancers", [])

        if len(balancers) == 0:
            return []

        arns = [balancer.get("LoadBalancerArn") for balancer in balancers]

        resource_tags = client.describe_tags(ResourceArns=arns)

        for tags in resource_tags:
            arn = tags.get("ResourceArn")
            origin_value = self._get_tag(tags, key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "elbv2"

    @property
    def _paginator(self):
        return "describe_load_balancers"

    @property
    def _filters(self):
        return {}



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("CloudWatch::Alarm")
class CloudWatchAlarmScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        alarms = []
        alarms.extend(item.get("CompositeAlarms", []))
        alarms.extend(item.get("MetricAlarms", []))

        for alarm in alarms:
            arn = alarm.get("AlarmArn", None)
            if arn is None:
                continue

            resource_tags = client.list_tags_for_resource(ResourceARN=arn)
            origin_value = self._get_tag(resource_tags, key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "cloudwatch"

    @property
    def _paginator(self):
        return "describe_alarms"

    @property
    def _filters(self):
        return {}



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("KMS")
class KMSScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        for key in item.get("Keys", []):
            key_arn = key.get("KeyArn", None)
            key_id = key.get("KeyId", None)
            if key_arn is None or key_id is None:
                continue

            resource_tags = client.list_resource_tags(KeyId=key_id)
            origin_value = self._get_tag(resource_tags, key)
            if overwrite or origin_value is None:
                resources.append((key_arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "kms"

    @property
    def _paginator(self):
        return "list_keys"

    @property
    def _filters(self):
        return {}

    @property
    def _tags_key(self) -> str:
        return "Items"



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("EMR::Cluster")
class EMRClusterScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        clusters = item.get("Clusters", [])
        for cluster in clusters:
            arn = cluster.get("ClusterArn", None)
            cluster_id = cluster.get("Id", None)
            if arn is None or cluster_id is None:
                continue

            detail = client.describe_cluster(ClusterId=cluster_id).get("Cluster", {})
            if overwrite or not self._has_tag(detail, key):
                resources.append((arn, self._get_tag(detail, key)))

        return resources

    @property
    def _client_name(self) -> str:
        return "emr"

    @property
    def _paginator(self):
        return "list_clusters"

    @property
    def _filters(self):
        return {"ClusterStates": ["STARTING", "BOOTSTRAPPING", "RUNNING", "WAITING"]}



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("RDS::Cluster")
class RDSClusterScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        clusters = item.get("DBClusters", [])
        for cluster in clusters:
            if overwrite is True or not self._has_tag(cluster, key):
                resources.append((cluster.get("DBClusterArn"), self._get_tag(cluster, key)))

        return resources

    @property
    def _client_name(self) -> str:
        return "rds"

    @property
    def _paginator(self):
        return "describe_db_clusters"

    @property
    def _filters(self):
        return {}

    @property
    def _tags_key(self) -> str:
        return "TagList"



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Iterable, List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("OpenSearch")
class OpenSearchScanner(Scanner):
    def _list_resources(self, client: any) -> Iterable[dict]:
        domains = client.list_domain_names(EngineType="OpenSearch").get("DomainNames", [])
        if len(domains) == 0:
            return []

        return [client.describe_elasticsearch_domains(DomainNames=[d.get("DomainName") for d in domains])]

    # noinspection PyUnusedLocal, DuplicatedCode
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        domains = item.get("DomainStatusList", [])
        for domain in domains:
            arn = domain.get("ARN", None)
            if arn is None:
                continue

            resource_tags = client.list_tags(ARN=arn)
            origin_value = self._get_tag(resource_tags, key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "opensearch"

    @property
    def _tags_key(self) -> str:
        return "TagList"



# -*- coding: utf-8 -*-

import logging
import time
from multiprocessing import Pipe, Process, Semaphore
from typing import List, Optional

from awstt.worker import *


logger = logging.getLogger(__name__)

PROCESS_SCAN_CONCURRENCY = 5


def _scan_in_process(
    sema: Semaphore,
    pipe: Pipe,
    worker_name: str,
    tag_key: str,
    account_id: str,
    partition: str,
    regions: List[str],
    profile: Optional[str] = None,
    overwrite: Optional[bool] = False,
):
    output = []

    try:
        sema.acquire()

        logger.info(f"resource scan is starting - {worker_name}")
        process_scanner = Scanner.by_name(worker_name)(
            account_id,
            partition=partition,
            regions=regions,
            profile=profile,
        )
        logger.info(f"resource scan is working - {worker_name}@{process_scanner.available_regions}")

        output = process_scanner.execute(key=tag_key, overwrite=overwrite)

        logger.info(f"resource scan is finished - {worker_name}")
    except KeyboardInterrupt:
        logger.warning(f"resource scan is terminated - {worker_name}")
    except Exception as e:
        logger.error(f"resource scan is failed - {worker_name}, exception - {e}")
    finally:
        pipe.send(output)
        pipe.close()
        sema.release()


def execute(evt: any, _):
    start = time.time()

    sema = Semaphore(PROCESS_SCAN_CONCURRENCY)
    processes = []
    output_pipes = []

    try:
        account_id = detect_account_id(evt.get("profile", None))
        partition = detect_partition(evt.get("partition", None))
        regions = detect_regions(evt.get("regions", None))
        profile = evt.get("profile", None)
        tag_key = evt.get("tag_key")
        tag_value = evt.get("tag_value")
        overwrite = evt.get("overwrite", False)

        params = dict(
            account_id=account_id,
            partition=partition,
            regions=regions,
            profile=profile,
            tag_key=tag_key,
            tag_value=tag_value,
            overwrite=overwrite,
        )

        logger.info(f"executing with params: {params}")

        for name in Scanner.list_available():
            parent_conn, child_conn = Pipe()
            output_pipes.append(parent_conn)

            kwargs = {"sema": sema, "worker_name": name, "pipe": child_conn, **params}
            kwargs.pop("tag_value")

            p = Process(target=_scan_in_process, kwargs=kwargs)
            processes.append(p)
            p.start()

        # make sure that all processes have finished
        for process in processes:
            process.join()

        regional_resources = {}
        for pipe in output_pipes:
            outputs = pipe.recv()

            for item in outputs:
                regional_resources[item[0]] = regional_resources.get(item[0], []) + item[1]

        logger.info(f"tagger now is running...")
        final_tagger = Tagger(account_id, partition=partition, profile=profile)
        final_tagger.execute(key=tag_key, value=tag_value, regional_resources=regional_resources)
        logger.info(f"tagger execute finished")

        logger.info(f"execute finished, total time used - {round(time.time() - start, 2)}s")
    except KeyboardInterrupt:
        for process in processes:
            process.terminate()
    except Exception as e:
        logger.error(f"Failed to execute, exception - {e}")



# -*- coding: utf-8 -*-

from __future__ import annotations

import logging
from abc import ABC
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import boto3

from awstt.worker.registrable import Registrable
from awstt.worker.utils import detect_region


logger = logging.getLogger(__name__)


class Scanner(Registrable, ABC):
    def __init__(self, account_id: str, *, partition: str = "aws", regions: List[str], profile: Optional[str] = None):
        """
        Initialize a resource scanner for service

        :param account_id: The AWS account id used to scan
        :type account_id: str
        :param partition: The AWS partition name, can be one of `'aws' | 'aws-cn' | 'aws-us-gov'`, default to `'aws'`
        :type partition: str
        :param regions: The AWS regions to scan
        :type regions: list of str
        :param profile: The AWS credentials' profile name to used when execute as cli, default to `None`
        :type profile: str, optional
        """

        session = boto3.Session(profile_name=profile)
        if len(regions) == 0:
            regions = session.get_available_regions(self._client_name, partition_name=partition)

        self._session = session
        self._account_id = account_id
        self._partition = partition
        self._available_regions = regions
        self._clients: List[any] = [self._create_client(self._session, region=r) for r in regions]

    @property
    def available_regions(self) -> List[str]:
        return self._available_regions

    def execute(self, *, key: str, overwrite: Optional[bool] = False) -> List[Tuple[str, List[Dict[str, str]]]]:
        """
        Tag resources with key and value

        :param key: The key of tag
        :type key: str
        :param overwrite: Whether to overwrite when the key is existed, default to `False`
        :type overwrite: bool, optional
        :return: A dict contains regional resources, @see `:class:awstt.worker.scanner.TaggingResource`
        :rtype: Dict[str, List]
        """
        regional_resources = []

        for client in self._clients:
            region = detect_region(client.meta.region_name, self._partition)
            paginator = self._list_resources(client)
            for page in paginator:
                resources = self._get_resources_from_page(client, page, key, overwrite)

                regional_resources.append(
                    (
                        region,
                        [dict(arn=self._build_arn(client, res[0]), original_value=res[1]) for res in resources],
                    )
                )

        return regional_resources

    def _create_client(self, session: any, *, region: str) -> any:
        """
        A shortcut to create boto3 service client

        :param session: The boto3 session
        :param region: The AWS region of client to connect
        :return: A boto3 service client instance
        """

        # avoid s3 get_bucket_tagging takes too long time if no tags set
        config = boto3.session.Config(connect_timeout=30, retries={"max_attempts": 0})

        return session.client(self._client_name, config=config, region_name=detect_region(region, self._partition))

    def _has_tag(self, item: dict, key: str) -> bool:
        """
        A shortcut to check whether the dict item of tags contain the key

        The method **MUST** be overridden in inherited class if root element of the `item` is not a list

        :param item: A dict object contains tags
        :param key: The key to detect whether exists
        :return: True if tags contain the key, otherwise False
        """
        return any(tag["Key"] == key for tag in item.get(self._tags_key, []))

    def _get_tag(self, item: dict, key: str) -> Optional[str]:
        """
        A shortcut to get the value from the dict item of tags by the key

        The method **MUST** be overridden in inherited class iif root element of the `item` is not a list

        :param item: A dict object contains tags
        :param key: The key to get value
        :return: The tag value if the key exists, otherwise None
        """
        return next((tag["Value"] for tag in item.get(self._tags_key, []) if tag["Key"] == key), None)

    def _list_resources(self, client: any) -> Iterable[dict]:
        """
        A shortcut to list resources, used `paginator` as default

        This method **MUST** be overridden in inherited class if `paginator` is not supported to list resources

        :param client: the boto3 client to execute method
        :return: An iterable dict contains resources information
        """
        return client.get_paginator(self._paginator).paginate(**self._filters)

    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        """
        Get resources from a dict return from paginator resources, @see `:method:_paging_resources`

        :param client: the boto3 client to execute the paginator, can be used to get relative information
        :param item: A dict object return from paginator
        :param key: The key will be tag to, used to check whether resource tags contains the key when require overwrite
        :param overwrite: Whether to overwrite if the key already existed
        :return: A list of tuples contain resource arn and original tag value, for example:
        `[
            ("arn:aws:s3:us-east-1:1234567890:bucket/tag_existed", "original_value"),
            ("arn:aws:s3:us-east-1:1234567890:bucket/tag_not_existed", None),
        ]`
        """
        raise NotImplementedError

    def _build_arn(self, client: any, arn_or_id: str) -> str:
        """
        A shortcut to build AWS ARN string

        :param client: The boto3 client used to get the resource, contains partition and region information
        :param arn_or_id: The original string return from `_get_resources`, can be an ARN or resource id
        :return: AWS ARN string
        """
        if arn_or_id.lower().startswith("arn:"):
            return arn_or_id

        prefix = f"arn:{self._partition}:{self._client_name}:{detect_region(client.meta.region_name, self._partition)}"

        return f"{prefix}:{self._account_id}:{self._arn_resource_type}/{arn_or_id}"

    @property
    def _client_name(self) -> str:
        """
        The client name used to create boto3 client instance

        :return: A boto3 client name
        """
        raise NotImplementedError

    @property
    def _arn_resource_type(self) -> str:
        """
        The ARN resource type of current working resources

        The property will not be used when response of paginator already contains ARN

        :return: An ARN resource type string
        """
        raise NotImplementedError

    @property
    def _paginator(self) -> str:
        """
        The paginator name used to list resources

        The property may not be used when `_list_resources` is overridden

        :return: A boto3 client paginator name
        """
        raise NotImplementedError

    @property
    def _filters(self) -> Dict[str, Any]:
        """
        The filters used in pagination

        The property may not be used when `_list_resources` is overridden

        :return: A dict contains filters, default is empty
        """
        return {}

    @property
    def _tags_key(self) -> str:
        """
        The root element name of tags items, used in `_get_tag` and `_has_tag`

        The property may not affect when `_get_tag` and `_has_tag` methods are overridden in inherited class

        :return:
        """
        return "Tags"



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Iterable, List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("EC2::VPNGateway")
class VPNGatewayScanner(Scanner):
    def _list_resources(self, client: any) -> Iterable[dict]:
        return [client.describe_vpn_gateways()]

    # noinspection PyUnusedLocal
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        gateways = item.get("VpnGateways", [])
        for gateway in gateways:
            if overwrite is True or not self._has_tag(gateway, key):
                arn = self._build_arn(client, gateway.get("VpnGatewayId"))
                resources.append((arn, self._get_tag(gateway, key)))

        return resources

    @property
    def _client_name(self) -> str:
        return "ec2"

    @property
    def _arn_resource_type(self) -> str:
        return "vpn-gateway"



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Iterable, List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("S3::Bucket")
class S3BucketScanner(Scanner):
    def _list_resources(self, client: any) -> Iterable[dict]:
        return [client.list_buckets()]

    # noinspection PyUnusedLocal
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        buckets = item.get("Buckets", [])
        for bucket in buckets:
            name = bucket.get("Name", None)
            if name is None:
                continue

            try:
                resource_tags = client.get_bucket_tagging(Bucket=name)
            except:  # noqa E722
                resource_tags = {"TagSet": []}

            if overwrite is True or not self._has_tag(resource_tags, key):
                arn = self._build_arn(client, name)
                resources.append((arn, self._get_tag(resource_tags, key)))

        return resources

    def _build_arn(self, client: any, arn_or_id: str) -> str:
        return f"arn:aws:s3:::{arn_or_id}"

    @property
    def _client_name(self) -> str:
        return "s3"

    @property
    def _arn_resource_type(self) -> str:
        return "s3"

    @property
    def _tags_key(self) -> str:
        return "TagSet"



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Optional, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("CloudTrail")
class CloudTrailScanner(Scanner):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        account_id: str,
        *,
        partition: str = "aws",
        regions: Optional[List[str]] = None,
        profile: Optional[str] = None,
    ):
        """
        CloudTrail is global service, ensure just create one global client
        """
        super().__init__(account_id, profile=profile, regions=["global"])

    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        trails = item.get("Trails", [])
        # noinspection SpellCheckingInspection
        arns = [t.get("TrailARN") for t in trails]

        resources_tags = []
        for chunks in [arns[i : i + 20] for i in range(0, len(arns), 20)]:  # noqa: E203
            """list max 20 resources per action"""
            for page in client.get_paginator("list_tags").paginate(ResourceIdList=chunks):
                resources_tags.extend(page.get("ResourceTagList", []))

        for tags in resources_tags:
            arn = tags.get("ResourceId")
            origin_value = self._get_tag(tags, key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "cloudtrail"

    @property
    def _arn_resource_type(self) -> str:
        return "trail"

    @property
    def _paginator(self):
        return "list_trails"

    @property
    def _filters(self):
        return {}

    @property
    def _tags_key(self) -> str:
        return "TagsList"



# -*- coding: utf-8 -*-

from typing import List, Optional

import boto3


def detect_account_id(profile: Optional[str] = None) -> str:
    session = boto3.Session(profile_name=profile)

    return session.client("sts").get_caller_identity().get("Account")


def detect_regions(regions: Optional[str]) -> List[str]:
    if regions is None:
        return []

    return [r.lower().strip() for r in regions.split(",") if len(r.strip()) > 0]


def detect_partition(partition: Optional[str]) -> str:
    if partition is None:
        return "aws"

    if partition.lower() not in ["aws", "aws-cn", "aws-us-gov"]:
        raise NotImplementedError("partition not supported")

    return partition.lower()


def detect_region(region_name: str, partition: str) -> str:
    if "global" in region_name and partition == "aws":
        return "us-east-1"

    return region_name



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Optional, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("CloudFront")
class CloudFrontScanner(Scanner):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        account_id: str,
        *,
        partition: str = "aws",
        regions: Optional[List[str]] = None,
        profile: Optional[str] = None,
    ):
        """
        CloudFront is global service, ensure just create one global client
        """
        super().__init__(account_id, profile=profile, regions=["global"])

    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        distributions = item.get("DistributionList", {})
        for dist_item in distributions.get("Items", []):
            arn = dist_item.get("ARN", None)
            if arn is None:
                continue

            resource_tags = client.list_tags_for_resource(Resource=arn)
            origin_value = self._get_tag(resource_tags.get("Tags", {}), key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "cloudfront"

    @property
    def _paginator(self):
        return "list_distributions"

    @property
    def _filters(self):
        return {}

    @property
    def _tags_key(self) -> str:
        return "Items"



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import List, Tuple, Union

from awstt.worker.scanner import Scanner


@Scanner.register("DynamoDB")
class DynamoDBScanner(Scanner):
    def _get_resources_from_page(
        self, client: any, item: dict, key: str, overwrite: bool = False
    ) -> List[Tuple[str, Union[str, None]]]:
        resources = []

        tables = item.get("TableNames", [])

        for table in tables:
            arn = self._build_arn(client, table)

            resource_tags = {self._tags_key: []}
            for page in client.get_paginator("list_tags_of_resource").paginate(ResourceArn=arn):
                resource_tags[self._tags_key].extend(page.get(self._tags_key, []))

            origin_value = self._get_tag(resource_tags, key)
            if overwrite or origin_value is None:
                resources.append((arn, origin_value))

        return resources

    @property
    def _client_name(self) -> str:
        return "dynamodb"

    @property
    def _arn_resource_type(self) -> str:
        return "table"

    @property
    def _paginator(self):
        return "list_tables"

    @property
    def _filters(self):
        return {}



# -*- coding: utf-8 -*-

import logging
import time
from typing import Dict, List, NotRequired, Optional, TypedDict

import boto3

from awstt.worker.utils import detect_region


logger = logging.getLogger(__name__)

RESOURCES_PER_REQUEST = 20  # The max limited to tag resources per request


class TaggingResource(TypedDict):
    arn: str
    original_value: NotRequired[str]


class Tagger(object):
    def __init__(self, account_id: str, *, partition: str = "aws", profile: Optional[str] = None):
        """
        Initialize a resource tagger to tag resources

        :param account_id: The AWS account id used to tag resources
        :type account_id: str
        :param partition: The AWS partition name, can be one of `'aws' | 'aws-cn' | 'aws-us-gov'`, default to `'aws'`
        :type partition: str
        :param profile: The profile name of AWS credentials used when execute as cli, default to `None`
        :type profile: str, optional
        """

        self._account_id = account_id
        self._partition = partition
        self._profile = profile

    def execute(self, *, key: str, value: str, regional_resources: Dict[str, List[TaggingResource]]):
        """
        Tag resources use ResourceGroupsTaggingApi

        :param regional_resources: The regional resources to be tagging
        :type regional_resources: Dict[str, List[TaggingResource]]
        :param key: The key of tag
        :type key: str
        :param value: The value of tag
        :type value: str
        """
        for region, resources in regional_resources.items():
            session = boto3.Session(profile_name=self._profile)
            # noinspection SpellCheckingInspection
            tagger = session.client(
                "resourcegroupstaggingapi",
                region_name=detect_region(region, self._partition),
            )

            if len(resources) == 0:
                continue

            for chunks in [
                resources[i : i + RESOURCES_PER_REQUEST]  # noqa: E203
                for i in range(0, len(resources), RESOURCES_PER_REQUEST)
            ]:
                """tag max 20 resources per action"""
                resp = tagger.tag_resources(ResourceARNList=[c.get("arn") for c in chunks], Tags={key: value})

                for res in chunks:
                    arn = res.get("arn")
                    original_value = res.get("original_value", None)

                    logger.info(
                        "resource tagged - %s",
                        dict(
                            resource=arn,
                            status=arn not in resp,
                            original_value=original_value,
                            error=resp.get(arn, {}).get("ErrorMessage", None),
                        ),
                    )

                time.sleep(1)  # avoid throttling
